from typing import List, Tuple, Optional

def safely_unquote_auth_item(string: str) -> str: ...
def safely_unquote_path(string: str) -> str: ...
def safely_unquote_query_item(string: str) -> str: ...
def safely_unquote_fragment(string: str) -> str: ...
def safely_unquote_qsl(
    qsl: List[Tuple[str, Optional[str]]]
) -> List[Tuple[str, Optional[str]]]: ...
def safely_quote(string: str) -> str: ...
def safely_quote_qsl(
    qsl: List[Tuple[str, Optional[str]]]
) -> List[Tuple[str, Optional[str]]]: ...
def upper_quoted(string: str) -> str: ...
